__author__ = "Simon Nilsson"

import warnings

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
import functools
import multiprocessing
import os
import platform
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import shutil

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_keys_exist_in_dict,
                                check_if_valid_rgb_tuple, check_int,
                                check_valid_boolean, check_valid_extension,
                                check_valid_tuple,
                                check_video_and_data_frm_count_align)
from simba.utils.data import plug_holes_shortest_bout
from simba.utils.enums import TextOptions
from simba.utils.printing import stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    create_directory, find_core_cnt,
                                    get_fn_ext, get_video_meta_data, read_df,
                                    read_pickle, write_df)


def _validation_video_mp(data: pd.DataFrame,
                        bp_dict: dict,
                        video_save_dir: str,
                        video_path: str,
                        text_thickness: int,
                        text_opacity: float,
                        font_size: int,
                        text_spacing: int,
                        circle_size: int,
                        show_pose: bool,
                        show_animal_names: bool,
                        gantt_setting: Union[int, None],
                        final_gantt: Optional[np.ndarray],
                        clf_data: np.ndarray,
                        clrs: List[List],
                        clf_name: str,
                        bouts_df: pd.DataFrame):

    def _put_text(img: np.ndarray,
                  text: str,
                  pos: Tuple[int, int],
                  font_size: int,
                  font_thickness: Optional[int] = 2,
                  font: Optional[int] = cv2.FONT_HERSHEY_DUPLEX,
                  text_color: Optional[Tuple[int, int, int]] = (255, 255, 255),
                  text_color_bg: Optional[Tuple[int, int, int]] = (0, 0, 0),
                  text_bg_alpha: float = 0.8):

        x, y = pos
        text_size, px_buffer = cv2.getTextSize(text, font, font_size, font_thickness)
        w, h = text_size
        overlay, output = img.copy(), img.copy()
        cv2.rectangle(overlay, (x, y-h), (x + w, y + px_buffer), text_color_bg, -1)
        cv2.addWeighted(overlay, text_bg_alpha, output, 1 - text_bg_alpha, 0, output)
        cv2.putText(output, text, (x, y), font, font_size, text_color, font_thickness)
        return output


    def _create_gantt(bouts_df: pd.DataFrame, clf_name: str, image_index: int, fps: int):

        fig, ax = plt.subplots(figsize=(final_gantt.shape[1] / dpi, final_gantt.shape[0] / dpi))
        matplotlib.font_manager._get_font.cache_clear()
        relRows = bouts_df.loc[bouts_df["End_frame"] <= image_index]
        for i, event in enumerate(relRows.groupby("Event")):
            data_event = event[1][["Start_time", "Bout_time"]]
            ax.broken_barh(data_event.values, (4, 4), facecolors="red")
        xLength = (round(image_index / fps)) + 1
        if xLength < 10:
            xLength = 10

        ax.set_xlim(0, xLength)
        ax.set_ylim([0, 12])
        ax.set_xlabel("Session (s)", fontsize=12)
        ax.set_ylabel(clf_name, fontsize=12)
        ax.set_title(f"{clf_name} GANTT CHART", fontsize=12)
        ax.set_yticks([])
        ax.yaxis.set_ticklabels([])
        ax.yaxis.grid(True)
        canvas = FigureCanvas(fig)
        canvas.draw()
        img = np.array(np.uint8(np.array(canvas.renderer._renderer)))[:, :, :3]
        plt.close(fig)
        return img

    dpi = plt.rcParams["figure.dpi"]
    fourcc, font = cv2.VideoWriter_fourcc(*"mp4v"), cv2.FONT_HERSHEY_DUPLEX
    cap = cv2.VideoCapture(video_path)
    video_meta_data = get_video_meta_data(video_path=video_path)
    batch_id, batch_data = data[0], data[1]
    start_frm, current_frm, end_frm = batch_data.index[0], batch_data.index[0], batch_data.index[-1]
    video_save_path = os.path.join(video_save_dir, f"{batch_id}.mp4")
    if gantt_setting is not None:
        video_size = (int(video_meta_data["width"] + final_gantt.shape[1]), int(video_meta_data["height"]))
        writer = cv2.VideoWriter(video_save_path, fourcc, video_meta_data["fps"], video_size)
    else:
        video_size = (int(video_meta_data["width"]), int(video_meta_data["height"]))
        writer = cv2.VideoWriter(video_save_path, fourcc, video_meta_data["fps"], video_size)
    cap.set(1, start_frm)
    while (current_frm <= end_frm) & (current_frm <= video_meta_data["frame_count"]):
        clf_frm_cnt = np.sum(clf_data[0:current_frm])
        ret, img = cap.read()
        if ret:
            if show_pose:
                for animal_cnt, (animal_name, animal_data) in enumerate(bp_dict.items()):
                    for bp_cnt, bp in enumerate(range(len(animal_data["X_bps"]))):
                        x_header, y_header = (animal_data["X_bps"][bp], animal_data["Y_bps"][bp])
                        animal_cords = tuple(batch_data.loc[current_frm, [x_header, y_header]])
                        cv2.circle(img, (int(animal_cords[0]), int(animal_cords[1])), circle_size, clrs[animal_cnt][bp_cnt], -1)
            if show_animal_names:
                for animal_cnt, (animal_name, animal_data) in enumerate(bp_dict.items()):
                    x_header, y_header = (animal_data["X_bps"][0], animal_data["Y_bps"][0],)
                    animal_cords = tuple(batch_data.loc[current_frm, [x_header, y_header]])
                    cv2.putText(img, animal_name, (int(animal_cords[0]), int(animal_cords[1])), font, font_size, clrs[animal_cnt][0], text_thickness)

            target_timer = round((1 / video_meta_data["fps"]) * clf_frm_cnt, 2)
            img = _put_text(img=img, text="BEHAVIOR TIMER:", pos=(TextOptions.BORDER_BUFFER_Y.value, text_spacing), font_size=font_size, font_thickness=TextOptions.TEXT_THICKNESS.value)
            addSpacer = 2
            img = _put_text(img=img, text=f"{clf_name} {target_timer}s", pos=(TextOptions.BORDER_BUFFER_Y.value, text_spacing * addSpacer), font_size=font_size, font_thickness=TextOptions.TEXT_THICKNESS.value, text_bg_alpha=text_opacity)
            addSpacer += 1
            img = _put_text(img=img, text="ENSEMBLE PREDICTION:", pos=(TextOptions.BORDER_BUFFER_Y.value, text_spacing * addSpacer), font_size=font_size, font_thickness=TextOptions.TEXT_THICKNESS.value, text_bg_alpha=text_opacity)
            addSpacer += 1
            if clf_data[current_frm] == 1:
                img = _put_text(img=img, text=clf_name, pos=(TextOptions.BORDER_BUFFER_Y.value, text_spacing * addSpacer), font_size=font_size, font_thickness=TextOptions.TEXT_THICKNESS.value, text_color=TextOptions.COLOR.value, text_bg_alpha=text_opacity)
            addSpacer += 1
            if gantt_setting == 1:
                img = np.concatenate((img, final_gantt), axis=1)
            elif gantt_setting == 2:
                gantt_img = _create_gantt(bouts_df, clf_name, current_frm, video_meta_data["fps"])
                img = np.concatenate((img, gantt_img), axis=1)
            img = cv2.resize(img, video_size, interpolation=cv2.INTER_LINEAR)
            writer.write(np.uint8(img))
            current_frm += 1
            print(f"Multi-processing video frame {current_frm} on core {batch_id}...")
    cap.release()
    writer.release()
    return batch_id


class ValidateModelOneVideoMultiprocess(ConfigReader, PlottingMixin, TrainModelMixin):
    """
    Create classifier validation video for a single input video. Results are stored in the
    ``project_folder/frames/output/validation`` directory.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str feature_file_path: path to SimBA file (parquet or CSV) containing pose-estimation and feature fields.
    :param str model_path: path to pickled classifier object
    :param float discrimination_threshold: classification threshold.
    :param int shortest_bout: Allowed classified bout length expressed in milliseconds. E.g., `1000` will shift frames classified
        as containing the behavior, but occuring in a bout shorter than `1000`, from `target present to `target absent`.
    :param str create_gantt:
        If SimBA should create gantt charts alongside the validation video. OPTIONS: 'None', 'Gantt chart: final frame only (slightly faster)',
        'Gantt chart: video'.
    :param dict settings: User style settings for video. E.g., {'pose': True, 'animal_names': True, 'styles': None}
    :param int cores: Number of cores to use.

    :example:
    >>> test = ValidateModelOneVideoMultiprocess(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
    >>>                              feature_file_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/features_extracted/Together_1.csv',
    >>>                              model_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/models/generated_models/Attack.sav',
    >>>                              discrimination_threshold=0.6,
    >>>                              shortest_bout=50,
    >>>                              cores=6,
    >>>                              settings={'pose': True, 'animal_names': True, 'styles': None},
    >>>                              create_gantt=None)
    >>> test.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 feature_path: Union[str, os.PathLike],
                 model_path: Union[str, os.PathLike],
                 show_pose: bool = True,
                 show_animal_names: bool = False,
                 font_size: Optional[bool] = None,
                 circle_size: Optional[int] = None,
                 text_spacing: Optional[int] = None,
                 text_thickness: Optional[int] = None,
                 text_opacity: Optional[float] = None,
                 discrimination_threshold: float = 0.0,
                 shortest_bout: int = 0.0,
                 core_cnt: int = -1,
                 create_gantt: Optional[Union[None, int]] = None):


        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        TrainModelMixin.__init__(self)
        check_file_exist_and_readable(file_path=config_path)
        check_file_exist_and_readable(file_path=feature_path)
        check_file_exist_and_readable(file_path=model_path)
        check_valid_boolean(value=[show_pose], source=f'{self.__class__.__name__} show_pose', raise_error=True)
        check_valid_boolean(value=[show_animal_names], source=f'{self.__class__.__name__} show_animal_names', raise_error=True)
        check_int(name=f"{self.__class__.__name__} core_cnt", value=core_cnt, min_value=-1, unaccepted_vals=[0])
        if font_size is not None: check_int(name=f'{self.__class__.__name__} font_size', value=font_size)
        if circle_size is not None: check_int(name=f'{self.__class__.__name__} circle_size', value=circle_size)
        if text_spacing is not None: check_int(name=f'{self.__class__.__name__} text_spacing', value=text_spacing)
        if text_opacity is not None: check_float(name=f'{self.__class__.__name__} text_opacity', value=text_opacity, min_value=0.1)
        if text_thickness is not None: check_float(name=f'{self.__class__.__name__} text_thickness', value=text_thickness, min_value=0.1)
        check_float(name=f"{self.__class__.__name__} discrimination_threshold", value=discrimination_threshold, min_value=0, max_value=1.0)
        check_int(name=f"{self.__class__.__name__} shortest_bout", value=shortest_bout, min_value=0)
        if create_gantt is not None:
            check_int(name=f"{self.__class__.__name__} create gantt", value=create_gantt, max_value=2, min_value=1)
        if not os.path.exists(self.single_validation_video_save_dir):
            os.makedirs(self.single_validation_video_save_dir)
        _, self.feature_filename, ext = get_fn_ext(feature_path)
        self.video_path = self.find_video_of_file(self.video_dir, self.feature_filename)
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        self.clf_name, self.feature_file_path = (os.path.basename(model_path).replace(".sav", ""), feature_path)
        self.vid_output_path = os.path.join(self.single_validation_video_save_dir, f"{self.feature_filename} {self.clf_name}.mp4")
        self.clf_data_save_path = os.path.join(self.clf_data_validation_dir, f"{self.feature_filename }.csv")
        self.show_pose, self.show_animal_names = show_pose, show_animal_names
        self.font_size, self.circle_size, self.text_spacing = font_size, circle_size, text_spacing
        self.text_opacity, self.text_thickness = text_opacity, text_thickness
        self.clf = read_pickle(data_path=model_path, verbose=True)
        self.data_df = read_df(feature_path, self.file_type)
        self.x_df = self.drop_bp_cords(df=self.data_df)
        self.discrimination_threshold, self.shortest_bout, self.create_gantt = float(discrimination_threshold), shortest_bout, create_gantt
        check_video_and_data_frm_count_align(video=self.video_path, data=self.data_df, name=self.feature_filename, raise_error=False)
        self.core_cnt = find_core_cnt()[0] if core_cnt == -1 or core_cnt > find_core_cnt()[0] else core_cnt
        self.temp_dir = os.path.join(self.single_validation_video_save_dir, "temp")
        self.video_save_path = os.path.join(self.single_validation_video_save_dir, f"{self.feature_filename}.mp4")
        create_directory(paths=self.temp_dir, overwrite=True)
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)

    def _get_styles(self):
        self.video_text_thickness = TextOptions.TEXT_THICKNESS.value if self.text_thickness is None else int(max(self.text_thickness, 1))
        longest_str = str(max(['TIMERS:', 'ENSEMBLE PREDICTION:'] + self.clf_names, key=len))
        optimal_font_size, _, optimal_spacing_scale = self.get_optimal_font_scales(text=longest_str, accepted_px_width=int(self.video_meta_data["width"] / 3), accepted_px_height=int(self.video_meta_data["height"] / 10), text_thickness=self.video_text_thickness)
        optimal_circle_size = self.get_optimal_circle_size(frame_size=(self.video_meta_data["width"], self.video_meta_data["height"]), circle_frame_ratio=100)
        self.video_circle_size = optimal_circle_size if self.circle_size is None else int(self.circle_size)
        self.video_font_size = optimal_font_size if self.font_size is None else self.font_size
        self.video_space_size = optimal_spacing_scale if self.text_spacing is None else int(max(self.text_spacing, 1))
        self.video_text_opacity = 0.8 if self.text_opacity is None else float(self.text_opacity)

    def run(self):
        self.prob_col_name = f"Probability_{self.clf_name}"
        self.data_df[self.prob_col_name] = self.clf_predict_proba(clf=self.clf, x_df=self.x_df, model_name=self.clf_name, data_path=self.feature_file_path)
        self.data_df[self.clf_name] = np.where(self.data_df[self.prob_col_name] > self.discrimination_threshold, 1, 0)
        if self.shortest_bout > 1:
            self.data_df = plug_holes_shortest_bout(data_df=self.data_df, clf_name=self.clf_name, fps=self.video_meta_data['fps'], shortest_bout=self.shortest_bout)
        _ = write_df(df=self.data_df, file_type=self.file_type, save_path=self.clf_data_save_path)
        print(f"Predictions created for video {self.feature_filename}...")
        if self.create_gantt is not None:
            self.bouts_df = self.get_bouts_for_gantt(data_df=self.data_df, clf_name=self.clf_name, fps=self.video_meta_data['fps'])
            self.final_gantt_img = self.create_gantt_img(self.bouts_df,self.clf_name,len(self.data_df), self.video_meta_data['fps'],"Behavior gantt chart (entire session)")
            self.final_gantt_img = self.resize_gantt(self.final_gantt_img, self.video_meta_data["height"])
        else:
            self.bouts_df, self.final_gantt_img = None, None
        self._get_styles()


        self.data_df = self.data_df.head(min(len(self.data_df), self.video_meta_data["frame_count"]))
        data = np.array_split(self.data_df, self.core_cnt)
        data = [(i, j) for i, j in enumerate(data)]

        with multiprocessing.Pool(self.core_cnt, maxtasksperchild=self.maxtasksperchild) as pool:
            constants = functools.partial(_validation_video_mp,
                                          bp_dict=self.animal_bp_dict,
                                          video_save_dir=self.temp_dir,
                                          text_thickness=self.video_text_thickness,
                                          text_opacity=self.video_text_opacity,
                                          font_size=self.video_font_size,
                                          text_spacing=self.video_space_size,
                                          circle_size=self.video_circle_size,
                                          video_path=self.video_path,
                                          show_pose=self.show_pose,
                                          show_animal_names=self.show_animal_names,
                                          gantt_setting=self.create_gantt,
                                          final_gantt=self.final_gantt_img,
                                          clf_data=self.data_df[self.clf_name].values,
                                          clrs=self.clr_lst,
                                          clf_name=self.clf_name,
                                          bouts_df=self.bouts_df)
            for cnt, result in enumerate(pool.imap(constants, data, chunksize=self.multiprocess_chunksize)):
                print(f"Image batch {result} complete, Video {self.feature_filename}...")
        pool.terminate()
        pool.join()
        concatenate_videos_in_folder(in_folder=self.temp_dir, save_path=self.video_save_path)
        self.timer.stop_timer()
        stdout_success(msg=f"Video complete, saved at {self.video_save_path}", elapsed_time=self.timer.elapsed_time_str)


# if __name__ == "__main__":
#     test = ValidateModelOneVideoMultiprocess(config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini",
#                                              feature_path=r"D:\troubleshooting\mitra\project_folder\csv\features_extracted\592_MA147_CNO1_0515.csv",
#                                              model_path=r"C:\troubleshooting\mitra\models\generated_models\grroming_undersample_2_1000\grooming.sav",
#                                              create_gantt=2,
#                                              show_pose=True,
#                                              show_animal_names=True,
#                                              core_cnt=5)
#     test.run()


#
# if __name__ == "__main__":
#     test = ValidateModelOneVideoMultiprocess(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                                     feature_file_path=r"C:\troubleshooting\mitra\project_folder\csv\features_extracted\844_MA131_gq_CNO_0624.csv",
#                                     model_path=r"C:\troubleshooting\mitra\models\validations\lay-on-belly_1\lay-on-belly.sav",
#                                     discrimination_threshold=0.35,
#                                     shortest_bout=200,
#                                     cores=-1,
#                                     settings={'pose': True, 'animal_names': False, 'styles': None},
#                                     create_gantt=2)
#     test.run()





# test = ValidateModelOneVideoMultiprocess(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/project_config.ini',
#                              feature_file_path='/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/csv/features_extracted/SI_DAY3_308_CD1_PRESENT.csv',
#                              model_path='/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/models/generated_models/Running.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              cores=6,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt=None)
# test.run()

# test = ValidateModelOneVideoMultiprocess(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                              feature_file_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/features_extracted/Together_1.csv',
#                              model_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/models/generated_models/Attack.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              cores=6,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt=None)
# test.run()



__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
import platform
from typing import List, Union

import cv2
import numpy as np
import pandas as pd
from numba import jit, prange

import simba.mixins.plotting_mixin
from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_filepaths_in_iterable_exist, check_int, check_str,
    check_valid_boolean, check_valid_dataframe, check_valid_dict)
from simba.utils.enums import Formats
from simba.utils.errors import InvalidInputError, NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, get_fn_ext, read_df,
                                    remove_a_folder)


def _heatmap_multiprocessor(data: np.array,
                            video_setting: bool,
                            frame_setting: bool,
                            video_temp_dir: str,
                            video_name: str,
                            frame_dir: str,
                            fps: int,
                            style_attr: dict,
                            max_scale: float,
                            clf_name: str,
                            aspect_ratio: float,
                            size: tuple,
                            make_clf_heatmap_plot: simba.mixins.plotting_mixin.PlottingMixin.make_clf_heatmap_plot):

    batch, frm_ids, data = data[0], data[1], data[2]
    if video_setting:
        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        video_save_path = os.path.join(video_temp_dir, f"{batch}.mp4")
        video_writer = cv2.VideoWriter(video_save_path, fourcc, fps, size)
    for frm_idx in range(data.shape[0]):
        frame_id, frm_data = int(frm_ids[frm_idx]), data[frm_idx]
        img = make_clf_heatmap_plot(frm_data=frm_data,
                                    max_scale=max_scale,
                                    palette=style_attr["palette"],
                                    aspect_ratio=aspect_ratio,
                                    shading=style_attr["shading"],
                                    clf_name=clf_name,
                                    img_size=size)

        print(f"Heatmap frame created: {frame_id + 1}, Video: {video_name}, Processing core: {batch+1}")
        if video_setting:
            video_writer.write(img)
        if frame_setting:
            file_path = os.path.join(frame_dir, f"{frame_id}.png")
            cv2.imwrite(file_path, img)

    if video_setting:
        video_writer.release()

    return batch


class HeatMapperClfMultiprocess(ConfigReader, PlottingMixin):
    """
    Create heatmaps representing the locations of the classified behavior.

    .. note::
       `GitHub visualizations tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

        .. image:: _static/img/heatmap.png
           :width: 500
           :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param bool final_img_setting: If True, then create a single image representing the last frame of the input video
    :param bool video_setting: If True, then create a video of heatmaps.
    :param bool frame_setting: If True, then create individual heatmap frames.
    :param int bin_size: The rectangular size of each heatmap location in millimeters. For example, `50` will divide the video into
        5 centimeter rectangular spatial bins.
    :param str palette: Heatmap pallette. Eg. 'jet', 'magma', 'inferno','plasma', 'viridis', 'gnuplot2'
    :param str bodypart: The name of the body-part used to infer the location of the classified behavior
    :param str clf_name: The name of the classified behavior.
    :param int or 'auto' max_scale: The max value in the heatmap in seconds. E.g., with a value of `10`, if the classified behavior has occured
        >= 10 within a rectangular bins, it will be filled with the same color.
    :param int core_cnt: Number of cores to use.

    :example I:
    >>> heat_mapper_clf = HeatMapperClfMultiprocess(config_path='MyConfigPath', final_img_setting=False, video_setting=True, frame_setting=False, bin_size=50, palette='jet', bodypart='Nose_1', clf_name='Attack', max_scale=20)
    >>> heat_mapper_clf.create_heatmaps()


    :example II:
    >>> test = HeatMapperClfMultiprocess(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini",
    >>>                          style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 50, 'max_scale': 'auto'},
    >>>                          final_img_setting=True,
    >>>                          video_setting=True,
    >>>                          frame_setting=True,
    >>>                          bodypart='Ear_left',
    >>>                          clf_name='straub_tail',
    >>>                          data_paths=[r"C:\troubleshooting\RAT_NOR\project_folder\csv\test\2022-06-20_NOB_DOT_4.csv"])
    >>> test.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 bodypart: str,
                 clf_name: str,
                 data_paths: List[str],
                 style_attr: dict,
                 final_img_setting: bool = True,
                 video_setting: bool = False,
                 frame_setting: bool = False,
                 core_cnt: int = -1):

        ConfigReader.__init__(self, config_path=config_path, create_logger=False)
        PlottingMixin.__init__(self)

        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        check_valid_boolean(value=[frame_setting, video_setting, final_img_setting], source=self.__class__.__name__)
        if (not frame_setting) and (not video_setting) and (not final_img_setting):
            raise NoSpecifiedOutputError(msg="Please choose to select either heatmap videos, frames, and/or final image.")
        check_filepaths_in_iterable_exist(file_paths=data_paths, name=f'{self.__class__.__name__} data_paths')
        check_str(name=f'{self.__class__.__name__} clf_name', value=clf_name)
        check_str(name=f'{self.__class__.__name__} bodypart', value=bodypart)
        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=-1, unaccepted_vals=[0])
        self.frame_setting, self.video_setting = frame_setting, video_setting
        self.final_img_setting, self.bp = final_img_setting, bodypart
        check_valid_dict(x=style_attr, required_keys=('max_scale', 'bin_size', 'shading', 'palette'))
        self.style_attr = style_attr
        self.bin_size, self.max_scale, self.palette, self.shading = (style_attr["bin_size"], style_attr["max_scale"], style_attr["palette"], style_attr["shading"])
        self.clf_name, self.data_paths = clf_name, data_paths
        self.core_cnt = [find_core_cnt()[0] if core_cnt == -1 or core_cnt > find_core_cnt()[0] else core_cnt][0]
        if not os.path.exists(self.heatmap_clf_location_dir):
            os.makedirs(self.heatmap_clf_location_dir)
        self.bp_lst = [f"{self.bp}_x", f"{self.bp}_y"]

    def __calculate_max_scale(self, clf_array: np.array):
        return np.round(np.max(np.max(clf_array[-1], axis=0)), 3)

    def run(self):
        print(f"Processing {len(self.data_paths)} video(s)...")
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            print(f'Plotting heatmap classification map for video {self.video_name}...')
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(video_name=self.video_name)
            self.width, self.height = int(self.video_info["Resolution_width"].values[0]), int(self.video_info["Resolution_height"].values[0])
            self.temp_folder = os.path.join(self.heatmap_clf_location_dir, "temp")
            self.frames_save_dir = os.path.join(self.heatmap_clf_location_dir, f"{self.video_name}_{self.clf_name}")
            if self.video_setting:
                if os.path.exists(self.temp_folder):
                    remove_a_folder(folder_dir=self.temp_folder)
                os.makedirs(self.temp_folder)
                self.save_video_path = os.path.join(self.heatmap_clf_location_dir, f"{self.video_name}_{self.clf_name}.mp4")
            if self.frame_setting:
                if os.path.exists(self.frames_save_dir):
                    remove_a_folder(folder_dir=self.frames_save_dir)
                os.makedirs(self.frames_save_dir)
            self.data_df = read_df(file_path=file_path, file_type=self.file_type)
            check_valid_dataframe(df=self.data_df, required_fields=[self.clf_name] + self.bp_lst, valid_dtypes=Formats.NUMERIC_DTYPES.value)
            bp_data = self.data_df[self.bp_lst].values.astype(np.int32)
            clf_data = self.data_df[self.clf_name].values.astype(np.int32)
            if len(np.unique(clf_data)) == 1:
                raise InvalidInputError(msg=f'Cannot plot heatmap for behavior {self.clf_name} in video {self.video_name}. The behavior is classified as {np.unique(clf_data)} in every single frame.')
            grid, aspect_ratio = GeometryMixin.bucket_img_into_grid_square(img_size=(self.width, self.height), bucket_grid_size_mm=self.bin_size, px_per_mm=self.px_per_mm, add_correction=False, verbose=False)
            clf_data = GeometryMixin().cumsum_bool_geometries(data=bp_data, geometries=grid, bool_data=clf_data, fps=self.fps, verbose=False)
            if self.max_scale == "auto":
                self.max_scale = max(1, self.__calculate_max_scale(clf_array=clf_data))
            if self.final_img_setting:
                file_name = os.path.join(self.heatmap_clf_location_dir, f"{self.video_name}_{self.clf_name}_final_frm.png")
                self.make_location_heatmap_plot(frm_data=clf_data[-1:, :, :][0],
                                                max_scale=self.max_scale,
                                                palette=self.palette,
                                                aspect_ratio=aspect_ratio,
                                                file_name=file_name,
                                                shading=self.shading,
                                                img_size=(self.width, self.height))
                print(f"Final heatmap image saved at {file_name}.")

            if self.video_setting or self.frame_setting:
                frame_arrays = np.array_split(clf_data, self.core_cnt)
                frm_per_core_w_batch = []
                frm_cnt = 0
                for batch_cnt in range(len(frame_arrays)):
                    frm_range = np.arange(frm_cnt, frm_cnt+ frame_arrays[batch_cnt].shape[0])
                    frm_cnt += len(frm_range)
                    frm_per_core_w_batch.append((batch_cnt, frm_range, frame_arrays[batch_cnt]))
                del frame_arrays
                print(f"Creating heatmaps, multiprocessing (chunksize: {self.multiprocess_chunksize}, cores: {self.core_cnt})...")
                with multiprocessing.Pool(self.core_cnt, maxtasksperchild=self.maxtasksperchild) as pool:
                    constants = functools.partial(_heatmap_multiprocessor,
                                                  video_setting=self.video_setting,
                                                  frame_setting=self.frame_setting,
                                                  style_attr=self.style_attr,
                                                  fps=self.fps,
                                                  video_temp_dir=self.temp_folder,
                                                  frame_dir=self.frames_save_dir,
                                                  max_scale=self.max_scale,
                                                  aspect_ratio=aspect_ratio,
                                                  clf_name=self.clf_name,
                                                  size=(self.width, self.height),
                                                  video_name=self.video_name,
                                                  make_clf_heatmap_plot=self.make_clf_heatmap_plot)

                    for cnt, batch in enumerate(pool.imap(constants, frm_per_core_w_batch, chunksize=self.multiprocess_chunksize)):
                        print(f'Batch core {batch+1}/{self.core_cnt} complete (Video {self.video_name})... ')
                    pool.terminate()
                    pool.join()

                if self.video_setting:
                    print(f"Joining {self.video_name} multiprocessed video...")
                    concatenate_videos_in_folder(in_folder=self.temp_folder, save_path=self.save_video_path)

                video_timer.stop_timer()
                print(f"Heatmap video {self.video_name} complete, (elapsed time: {video_timer.elapsed_time_str}s) ...")

        self.timer.stop_timer()
        stdout_success(msg=f"Heatmap visualizations for {len(self.data_paths)} video(s) created in {self.heatmap_clf_location_dir} directory", elapsed_time=self.timer.elapsed_time_str)


# if __name__ == "__main__":
#     test = HeatMapperClfMultiprocess(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini",
#                              style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 50, 'max_scale': 'auto'},
#                              final_img_setting=True,
#                              video_setting=True,
#                              frame_setting=True,
#                              bodypart='Ear_left',
#                              clf_name='straub_tail',
#                              data_paths=[r"C:\troubleshooting\RAT_NOR\project_folder\csv\test\2022-06-20_NOB_DOT_4.csv"])
#     test.run()



# if __name__ == "__main__":
#     test = HeatMapperClfMultiprocess(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini",
#                          style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 50, 'max_scale': 'auto'},
#                          final_img_setting=True,
#                          video_setting=True,
#                          frame_setting=False,
#                          bodypart='Ear_left',
#                          clf_name='straub_tail',
#                          data_paths=[r"C:\troubleshooting\RAT_NOR\project_folder\csv\test\2022-06-20_NOB_DOT_4.csv"])
#     test.run()

# test = HeatMapperClfMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                      style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'},
#                      final_img_setting=False,
#                      video_setting=True,
#                      frame_setting=False,
#                      bodypart='Nose_1',
#                      clf_name='Attack',
#                                core_cnt=5,
#                      files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'])
# test.create_heatmaps()



import os
from tkinter import *
from typing import List, Optional, Union

from simba.data_processors.cue_light_movement_statistics import \
    CueLightMovementAnalyzer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        SimBADropDown)
from simba.utils.checks import check_if_dir_exists, check_int, check_valid_lst
from simba.utils.enums import Links
from simba.utils.errors import NoFilesFoundError, NoROIDataError
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    str_2_bool)


class CueLightMovementAnalyzerPopUp(ConfigReader, PopUpMixin):

    """
    :example:
    >>> CueLightMovementAnalyzerPopUp(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini", cue_light_names=['cl'])
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 cue_light_names: List[str],
                 data_dir: Optional[Union[str, os.PathLike]] = None):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=True)
        check_valid_lst(data=cue_light_names, source=self.__class__.__name__, valid_dtypes=(str,), min_len=1, raise_error=True)
        self.read_roi_data()
        self.cue_light_names = cue_light_names
        missing_cue_lights = [x for x in cue_light_names if x not in self.roi_names]
        if len(missing_cue_lights) > 0: raise NoROIDataError(msg=f'{len(missing_cue_lights)} cue-lights are not drawn in the SimBA project: cannot be found in the {self.roi_coordinates_path} file', source=self.__class__.__name__)

        if data_dir is None:
            self.data_dir = self.cue_lights_data_dir
        else:
            check_if_dir_exists(in_dir=data_dir)
            self.data_dir = data_dir

        self.data_paths = find_files_of_filetypes_in_directory(directory=self.data_dir, extensions=[f'.{self.file_type}'], raise_error=False, as_dict=True)
        file_cnt = len(list(self.data_paths.keys())) if type(self.data_paths) == dict else len(self.data_paths)
        if file_cnt == 0:
            raise NoFilesFoundError(msg=f'No data files exist in the {self.data_dir} directory: ANALYZE CUE LIGHT DATA before performing cue light movement statistics analysis.', source=self.__class__.__name__)

        PopUpMixin.__init__(self, size=(750, 300), title="CUE LIGHT MOVEMENT ANALYSIS", icon='analyze_green')

        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings', icon_link=Links.CUE_LIGHTS.value)
        self.bp_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=self.body_parts_lst, label=f'SELECT BODY-PART:', label_width=40, dropdown_width=30, value=self.body_parts_lst[0])
        self.pre_window_eb = Entry_Box(parent=self.settings_frm, fileDescription='PRE CUE LIGHT WINDOW (S): ', entry_box_width=30, value=0, validation='numeric', labelwidth="40", justify='center')
        self.post_window_eb = Entry_Box(parent=self.settings_frm, fileDescription='POST CUE LIGHT WINDOW (S): ', entry_box_width=30, value=0, validation='numeric', labelwidth="40", justify='center')

        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.bp_dropdown.grid(row=0, column=0, sticky=NW)
        self.pre_window_eb.grid(row=1, column=0, sticky=NW)
        self.post_window_eb.grid(row=2, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)

    def run(self):
        bp_name = self.bp_dropdown.get_value()
        pre_window = self.pre_window_eb.entry_get
        post_window = self.post_window_eb.entry_get
        check_int(name='PRE CUE LIGHT WINDOW (S): ', value=pre_window, min_value=0, raise_error=True)
        check_int(name='POST CUE LIGHT WINDOW (S): ', value=post_window, min_value=0, raise_error=True)

        analyzer = CueLightMovementAnalyzer(config_path=self.config_path, cue_light_names=self.cue_light_names, bp_name=bp_name, pre_window=int(pre_window), post_window=int(post_window))
        analyzer.run()
        analyzer.save()



#CueLightMovementAnalyzerPopUp(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini", cue_light_names=['cl'])






__author__ = "Simon Nilsson"

from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        SimbaButton)
from simba.utils.checks import check_str
from simba.utils.enums import Formats, Keys, Links
from simba.utils.read_write import archive_processed_files


class ArchiveProcessedFilesPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="ARCHIVE PROCESSED FILES", icon='archive')
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        archive_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header='ARCHIVE PROCESSED FILES', icon_name=Keys.DOCUMENTATION.value, icon_link=Links.SCENARIO_4.value)
        self.archive_eb = Entry_Box(archive_frm, "ARCHIVE DIRECTORY NAME:", "25")
        archive_btn = SimbaButton(parent=archive_frm, txt='RUN ARCHIVE', img='rocket', cmd=self.run)
        archive_frm.grid(row=0, column=0, sticky=NW)
        self.archive_eb.grid(row=0, column=0, sticky=NW)
        archive_btn.grid(row=1, column=0, sticky=NW)

    def run(self):
        archive_name = self.archive_eb.entry_get.strip()
        check_str(name="ARCHIVE NAME", value=archive_name, raise_error=True)
        archive_processed_files(config_path=self.config_path, archive_name=archive_name)



import os
import threading
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.ez_path_plot import EzPathPlot
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon,
                                        SimBADropDown, SimBALabel)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_valid_rgb_tuple, check_int, check_str)
from simba.utils.enums import Formats
from simba.utils.errors import NoFilesFoundError
from simba.utils.lookups import get_color_dict
from simba.utils.read_write import get_fn_ext, read_video_info


class QuickLineplotPopup(PopUpMixin, ConfigReader):
    def __init__(self,
                 config_path: Union[str, os.PathLike]):
        """
        :example:
        >>> _ = QuickLineplotPopup(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
        """

        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        if len(self.outlier_corrected_paths) == 0:
            raise NoFilesFoundError(msg=f"No data found in the {self.outlier_corrected_paths} directory. Place files in this directory to create quick path plots.", source=self.__class__.__name__)
        PopUpMixin.__init__(self, title="SIMPLE LINE PLOT", icon='path_2')
        self.video_filepaths = {get_fn_ext(filepath=i)[1]: i for i in self.outlier_corrected_paths}
        color_lst = list(get_color_dict().keys())
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header='SETTINGS', icon_name='settings')
        video_names = list(self.video_filepaths.keys())
        video_name_max_len = max(len(s) for s in video_names)

        self.video_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=video_names, label='VIDEO: ', label_width=25, dropdown_width=video_name_max_len, value=video_names[0])
        self.bp_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=self.body_parts_lst, label='BODY-PART: ', label_width=25, dropdown_width=video_name_max_len, value=self.body_parts_lst[0])
        self.background_color = SimBADropDown(parent=settings_frm, dropdown_options=color_lst, label="BACKGROUND COLOR: ", label_width=25, dropdown_width=video_name_max_len, value="White")
        self.line_color = SimBADropDown(parent=settings_frm, dropdown_options=color_lst, label="LINE COLOR: ", label_width=25, dropdown_width=video_name_max_len, value="Red")
        self.line_thickness = SimBADropDown(parent=settings_frm, dropdown_options=list(range(1, 11)), label="LINE THICKNESS: ", label_width=25, dropdown_width=video_name_max_len, value=1)
        self.circle_size = SimBADropDown(parent=settings_frm, dropdown_options=list(range(1, 11)), label="CIRCLE SIZE: ", label_width=25, dropdown_width=video_name_max_len, value=5)
        self.last_frm_only_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=["TRUE", "FALSE"], label="LAST FRAME ONLY: ", label_width=25, dropdown_width=video_name_max_len, value='FALSE')
        self.inst_lbl = SimBALabel(parent=settings_frm, txt="NOTE: For more complex path plots, faster, \n see 'CREATE PATH PLOTS' under the [VISUALIZATIONS] tab", txt_clr='green')

        settings_frm.grid(row=0, sticky=W)
        self.video_dropdown.grid(row=0, sticky=W)
        self.bp_dropdown.grid(row=2, sticky=W)
        self.background_color.grid(row=3, sticky=W)
        self.line_color.grid(row=4, sticky=W)
        self.line_thickness.grid(row=5, sticky=W)
        self.circle_size.grid(row=6, sticky=W)
        self.last_frm_only_dropdown.grid(row=7, sticky=W)
        self.inst_lbl.grid(row=8, sticky=W)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        video_name = self.video_dropdown.getChoices()
        data_path = self.video_filepaths[video_name]
        meta_data, _, fps = read_video_info(vid_info_df=self.video_info_df, video_name=video_name)
        size = (int(meta_data["Resolution_width"]), int(meta_data["Resolution_height"]))
        last_frm = self.last_frm_only_dropdown.getChoices()
        if last_frm == "TRUE":
            save_path = os.path.join(
                self.path_plot_dir, f"{video_name}_simple_path_plot.png"
            )
            last_frm = True
        else:
            save_path = os.path.join(
                self.path_plot_dir, f"{video_name}_simple_path_plot.mp4"
            )
            last_frm = False
        if not os.path.isdir(self.path_plot_dir):
            os.makedirs(self.path_plot_dir)
        background_color = get_color_dict()[self.background_color.getChoices()]
        line_color = get_color_dict()[self.line_color.getChoices()]
        line_thickness = self.line_thickness.getChoices()
        circle_size = self.circle_size.getChoices()
        bp = self.bp_dropdown.getChoices()
        check_int(
            name=f"{self.__class__.__name__} line_thickness",
            value=line_thickness,
            min_value=1,
        )
        check_int(
            name=f"{self.__class__.__name__} circle_size",
            value=circle_size,
            min_value=1,
        )
        check_if_valid_rgb_tuple(data=background_color)
        check_if_valid_rgb_tuple(data=line_color)
        check_str(name=f"{self.__class__.__name__} body-part", value=bp)
        plotter = EzPathPlot(
            data_path=data_path,
            size=size,
            fps=fps,
            body_part=bp,
            bg_color=background_color,
            line_color=line_color,
            line_thickness=int(line_thickness),
            circle_size=int(circle_size),
            last_frm_only=last_frm,
            save_path=save_path,
        )
        threading.Thread(target=plotter.run).start()



# _ = QuickLineplotPopup(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
# _ = QuickLineplotPopup(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# _ = QuickLineplotPopup(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import threading
from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.converters import dlc_to_labelme
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimBADropDown)
from simba.utils.checks import check_if_dir_exists
from simba.utils.read_write import str_2_bool


class DLCAnnotations2LabelMePopUp(PopUpMixin):

    """
    :example:
    >>> DLCAnnotations2LabelMePopUp()
    """
    def __init__(self):
        super().__init__(title="DLC ANNOTATIONS TO LABELME")
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.dlc_dir = FolderSelect(settings_frm, folderDescription="DLC ANNOTATIONS DIRECTORY:", lblwidth=35)
        self.labelme_dir = FolderSelect(settings_frm, folderDescription="LABELME SAVE DIR:", lblwidth=35)
        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="VERBOSE:", label_width=35, dropdown_width=40, value='TRUE')

        settings_frm.grid(row=0, column=0, sticky=NW)
        self.dlc_dir.grid(row=0, column=0, sticky=NW)
        self.labelme_dir.grid(row=1, column=0, sticky=NW)
        self.verbose_dropdown.grid(row=2, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()


    def run(self):
        dlc_dir = self.dlc_dir.folder_path
        labelme_dir = self.labelme_dir.folder_path
        verbose = str_2_bool(self.verbose_dropdown.get_value())

        check_if_dir_exists(in_dir=dlc_dir, source=f'{self.__class__.__name__} DLC ANNOTATIONS DIRECTORY')
        check_if_dir_exists(in_dir=labelme_dir, source=f'{self.__class__.__name__} LABELME SAVE DIR')

        thread = threading.Thread(target=dlc_to_labelme, kwargs={'dlc_dir': dlc_dir, 'save_dir': labelme_dir, 'verbose': verbose})
        thread.start()





__author__ = "Simon Nilsson"
__email__ = "sronilsson@gmail.com"

import os
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import shap
from sklearn.ensemble import RandomForestClassifier

from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_if_dir_exists, check_instance, check_int,
                                check_nvidea_gpu_available, check_str,
                                check_valid_array, check_valid_dataframe,
                                check_valid_lst)
from simba.utils.enums import Formats
from simba.utils.errors import FFMPEGCodecGPUError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import write_df
from simba.utils.warnings import NotEnoughDataWarning


def create_shap_log(rf_clf: Union[str, os.PathLike, RandomForestClassifier],
                    x: Union[pd.DataFrame, np.ndarray],
                    y: Union[pd.DataFrame, pd.Series, np.ndarray],
                    cnt_present: int,
                    cnt_absent: int,
                    x_names: Optional[List[str]] = None,
                    clf_name: Optional[str] = None,
                    save_dir: Optional[Union[str, os.PathLike]] = None,
                    verbose: Optional[bool] = True) -> Union[None, Tuple[pd.DataFrame, pd.DataFrame, int]]:
    """
    Computes SHAP (SHapley Additive exPlanations) values using a GPU for a RandomForestClassifier,
    based on specified counts of positive and negative samples, and optionally saves the results.

    .. image:: _static/img/create_shap_log_cuda.png
       :width: 500
       :align: center

    .. note::
       (i) The SHAP library has to be built from git repo rather than pip: `pip install git+https://github.com/slundberg/shap.git`.
       (ii) The scikit model cannot be built using max_depth > 31. You can set this in the SimBA config under [create ensemble settings][rf_max_depth], or `rf_max_depth` in the config CSV's.

    :param Union[str, os.PathLike, RandomForestClassifier] rf_clf: Trained RandomForestClassifier model or path to the saved model. Can be a string, os.PathLike object, or an instance of RandomForestClassifier.
    :param Union[pd.DataFrame, np.ndarray] x: Input features used for SHAP value computation. Can be a pandas DataFrame or numpy ndarray.
    :param Union[pd.DataFrame, pd.Series, np.ndarray] y:  Target labels corresponding to the input features. Can be a pandas DataFrame, pandas Series, or numpy ndarray with 0 and 1 values.
    :param int cnt_present: Number of positive samples (label=1) to include in the SHAP value computation.
    :param int cnt_absent: Number of negative samples (label=0) to include in the SHAP value computation.
    :param Optional[List[str]] x_names: Optional list of feature names corresponding to the columns in `x`. If `x` is a DataFrame, this is extracted automatically.
    :param Optional[str] clf_name: Optional name for the classifier, used in naming output files. If not provided, it is extracted from the `y` labels if possible.
    :param Optional[Union[str, os.PathLike]] save_dir:  Optional directory path where the SHAP values and corresponding raw features are saved as CSV files.
    :param Optional[bool] verbose: Optional boolean flag indicating whether to print progress messages. Defaults to True.
    :return Union[None, Tuple[pd.DataFrame, pd.DataFrame, int]]: If `save_dir` is None, returns a tuple containing:
                                                                - V: DataFrame with SHAP values, expected value, sum of SHAP values, prediction probability, and target labels.
                                                                - R: DataFrame containing the raw feature values for the selected samples.
                                                                - expected_value: The expected value from the SHAP explainer.
                                                                 If `save_dir` is provided, the function returns None and saves the output to CSV files in the specified directory.

    :example:
    >>> x = np.random.random((1000, 501)).astype(np.float32)
    >>> y = np.random.randint(0, 2, size=(len(x), 1)).astype(np.int32)
    >>> clf_names = [str(x) for x in range(501)]
    >>> results = create_shap_log(rf_clf=MODEL_PATH, x=x, y=y, cnt_present=int(i/2), cnt_absent=int(i/2), clf_name='TEST', x_names=clf_names, verbose=False)
    """

    timer = SimbaTimer(start=True)
    if verbose:
        print('Computing SHAP values (GPU)...')
    if not check_nvidea_gpu_available():
        raise FFMPEGCodecGPUError(msg="No GPU found (as evaluated by nvidea-smi returning None)",
                                  source=create_shap_log.__name__)
    check_instance(source=f'{create_shap_log.__name__} rf_clf', instance=rf_clf,
                   accepted_types=(str, RandomForestClassifier))
    if isinstance(rf_clf, (str, os.PathLike)):
        rf_clf = TrainModelMixin().read_pickle(file_path=rf_clf)
    check_instance(source=f'{create_shap_log.__name__} x', instance=x, accepted_types=(pd.DataFrame, np.ndarray))
    if isinstance(x, np.ndarray):
        check_valid_lst(data=x_names, source=f'{create_shap_log.__name__} x_names', valid_dtypes=(str,),
                        exact_len=x.shape[1])
        check_valid_array(data=x, source=f'{create_shap_log.__name__} x', accepted_ndims=[2, ],
                          accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    else:
        check_valid_dataframe(df=x, source=f'{create_shap_log.__name__} x',
                              valid_dtypes=Formats.NUMERIC_DTYPES.value)
        x_names = list(x.columns)
        x = x.values
    check_instance(source=f'{create_shap_log.__name__} y', instance=y,
                   accepted_types=(pd.DataFrame, np.ndarray, pd.Series))
    if isinstance(y, np.ndarray):
        check_str(name=f'{create_shap_log.__name__} clf_name', value=clf_name)
        y = y.flatten()
    elif isinstance(y, pd.Series):
        clf_name = y.name
        y = y.values.flatten()
    else:
        check_valid_dataframe(df=y, source=f'{create_shap_log.__name__} y',
                              valid_dtypes=Formats.NUMERIC_DTYPES.value, max_axis_1=1)
        clf_name = list(y.columns)[0]
        y = y.values.flatten()
    save_shap_path, save_raw_path = None, None
    if save_dir is not None:
        check_if_dir_exists(in_dir=save_dir)
        save_shap_path = os.path.join(save_dir, f"SHAP_values_{clf_name}.csv")
        save_raw_path = os.path.join(save_dir, f"RAW_SHAP_feature_values_{clf_name}.csv")
    check_valid_array(data=y, source=f'{create_shap_log.__name__} y', accepted_values=[0, 1])
    check_int(name=f'{create_shap_log.__name__} cnt_present', value=cnt_present, min_value=1)
    check_int(name=f'{create_shap_log.__name__} cnt_absent', value=cnt_absent, min_value=1)
    target_cnt = np.sum(y)
    absent_cnt = y.shape[0] - target_cnt

    if cnt_present > target_cnt:
        NotEnoughDataWarning(
            msg=f"Data contains {target_cnt} behavior-present annotations. This is less the number of frames you specified to calculate shap values for ({cnt_present}). SimBA will calculate shap scores for the {target_cnt} behavior-present frames available",
            source=create_shap_log.__name__)
        cnt_present = target_cnt
    if absent_cnt < cnt_absent:
        NotEnoughDataWarning(
            msg=f"Data contains {absent_cnt} behavior-absent annotations. This is less the number of frames you specified to calculate shap values for ({cnt_absent}). SimBA will calculate shap scores for the {absent_cnt} behavior-absent frames available",
            source=create_shap_log.__name__)
        cnt_absent = absent_cnt

    target_idx = np.argwhere(y == 1).flatten()
    absent_idx = np.argwhere(y == 0).flatten()
    target_idx = np.sort(np.random.choice(target_idx, cnt_present))
    absent_idx = np.sort(np.random.choice(absent_idx, cnt_absent))
    target_x = x[target_idx]
    absent_x = x[absent_idx]
    X = np.vstack([target_x, absent_x]).astype(np.float32)
    Y = np.hstack([np.ones(target_x.shape[0]), np.zeros(absent_x.shape[0])]).astype(np.int32)
    explainer = shap.explainers.GPUTree(model=rf_clf, data=None, model_output='raw', feature_names='tree_path_dependent')
    shap_values = explainer.shap_values(X, check_additivity=True)
    V = pd.DataFrame(shap_values[1], columns=x_names).astype(np.float32)
    sum = V.sum(axis=1)
    expected_value = explainer.expected_value[1]
    p = TrainModelMixin().clf_predict_proba(clf=rf_clf, x_df=X)

    V['EXPECTED_VALUE'] = expected_value.round(4)
    V['SUM'] = sum + V['EXPECTED_VALUE']
    V['PREDICTION_PROBABILITY'] = p.round(4)
    V['SUM'] = V['SUM'].round(4)
    V[clf_name] = Y
    x_idx = np.hstack([target_idx, absent_idx])
    R = pd.DataFrame(x[x_idx, :], columns=x_names)
    timer.stop_timer()
    if save_dir is None:
        if verbose:
            stdout_success(msg=f'Shap values compute complete (GPU) for {len(V)} observations.',  elapsed_time=timer.elapsed_time_str)
        return (V, R, expected_value)
    else:
        write_df(df=V, file_type='csv', save_path=save_shap_path)
        write_df(df=R, file_type='csv', save_path=save_raw_path)
        if verbose:
            stdout_success(msg=f'Shap values compute complete (GPU) for {len(V)} observations, and saved in {save_dir}',  elapsed_time=timer.elapsed_time_str)









import requests
import pandas as pd
import time
from concurrent.futures import ThreadPoolExecutor
import os
from zipfile import ZipFile

# === CONFIG ===
REPO = "sgoldenlab/simba"  # Repo: sgoldenlab/simba
GITHUB_TOKEN = os.getenv("GITHUB_TOKEN")  # Get the GitHub token from environment variables
PER_PAGE = 100
HEADERS = {"Accept": "application/vnd.github+json"}
if GITHUB_TOKEN:
    HEADERS["Authorization"] = f"token {GITHUB_TOKEN}"

# === PAGINATION HANDLER ===
def fetch_paginated(url, params=None):
    results = []
    page = 1
    while True:
        query = params.copy() if params else {}
        query.update({'per_page': PER_PAGE, 'page': page})
        response = requests.get(url, headers=HEADERS, params=query)
        if response.status_code != 200:
            print(f"Error: {response.status_code} - {response.text}")
            break
        page_data = response.json()
        if not page_data:
            break
        results.extend(page_data)
        page += 1
        time.sleep(0.1)
    return results

# === FETCH ISSUES ===
def fetch_all_issues():
    print("Fetching issues...")
    return fetch_paginated(f"https://api.github.com/repos/{REPO}/issues", {"state": "all"})

# === FETCH COMMENTER USERNAMES AND LAST COMMENT DATE ===
def get_comment_usernames_and_last_comment_date(issue_number):
    url = f"https://api.github.com/repos/{REPO}/issues/{issue_number}/comments"
    try:
        comments = fetch_paginated(url)
        usernames = {c["user"]["login"] for c in comments}
        last_comment_date = max(c["created_at"] for c in comments) if comments else None
        return sorted(usernames), last_comment_date
    except Exception as e:
        print(f"Failed on issue #{issue_number}: {e}")
        return [], None

# === MAIN FUNCTION ===
def get_issues_dataframe():
    issues = fetch_all_issues()
    print(f"Fetched {len(issues)} total issues.")

    # Remove pull requests
    issues = [i for i in issues if "pull_request" not in i]
    print(f"Filtered to {len(issues)} issues (excluding PRs).")

    # Fetch comment usernames and last comment date in parallel
    with ThreadPoolExecutor(max_workers=10) as executor:
        usernames_and_dates = list(executor.map(
            lambda i: get_comment_usernames_and_last_comment_date(i["number"]),
            issues
        ))

    data = []
    for issue, (commenters, last_comment_date) in zip(issues, usernames_and_dates):
        author = issue["user"]["login"]
        all_participants = set(commenters)
        all_participants.add(author)

        data.append({
            "issue_id": issue["number"],
            "author": author,
            "title": issue["title"],
            "commenter_usernames": sorted(commenters),
            "comment_count": issue.get("comments", 0),
            "unique_commenter_count": len(all_participants),
            "link": f"https://github.com/{REPO}/issues/{issue['number']}",
            "issue_post_date": issue["created_at"],
            "last_comment_date": last_comment_date,
        })

    df = pd.DataFrame(data)

    # Save DataFrame to CSV
    df.to_csv("/tmp/github_issues_report.csv", index=False)

    # Zip the CSV file
    with ZipFile("/tmp/github_issues_report.zip", 'w') as zipf:
        zipf.write("/tmp/github_issues_report.csv", os.path.basename("/tmp/github_issues_report.csv"))

    print("CSV and Zip file generated successfully!")

if __name__ == "__main__":
    get_issues_dataframe()



import argparse
import os
from itertools import groupby
from typing import Optional, Union

import numpy as np
import pandas as pd

from simba.utils.checks import (check_float, check_if_dir_exists, check_str,
                                check_valid_dataframe)
from simba.utils.data import detect_bouts
from simba.utils.enums import Formats
from simba.utils.printing import SimbaTimer
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    seconds_to_timestamp, write_df)

pd.options.mode.chained_assignment = None

COLUMN_NAMES = ['EVENT', 'START TIME (S)', 'END TIME (S)', 'START FRAME', 'END FRAME', 'DURATION (S)']
OUT_COL_NAMES= ['VIDEO', 'BODY-PART', 'EVENT', 'START TIME (S)', 'END TIME (S)', 'START FRAME', 'END FRAME', 'DURATION (S)']

class LightDarkBoxAnalyzer():
    """
    Perform light–dark box analysis using DeepLabCut pose estimation data.

    .. note::
       If the specified body-part is detected below the specified threshold, then the animal is the dark-box. Otherwise its in the light-box.

    .. seealso::
       For light/dark box data plotting, see :func:`simba.plotting.light_dark_box_plotter.LightDarkBoxPlotter`.

    This class analyzes animal transitions between light and dark compartments  in a light–dark box behavioral test based on the availability of pose-estimation data. It assumes that if pose-estimation for a specified body part is available,
    the animal is in the light box; otherwise, the animal is in the dark box. It detects bouts (continuous time segments) for each condition and saves the bout-level data to a CSV file.

    :param Union[str, os.PathLike] data_dir: Directory containing DeepLabCut CSV files with pose-estimation data.
    :param Union[str, os.PathLike] save_path: Full path to save the resulting CSV file with light/dark bouts.
    :param str body_part: The name of the body part used to infer the animal's presence.
    :param Union[int, float] fps: Frames per second of the video recordings.
    :param float threshold: Value between 0 and 1. If below this value, animal is in dark box. If above, animal is in light box.

    :example:
    >>> python light_dark_box_analyzer.py --data_dir 'D:\light_dark_box\project_folder\csv\input_csv' --save_path "D:\light_dark_box\project_folder\csv\results\light_dark_data.csv" --body_part nose --fps 29 --threshold 0.01
    >>> analyzer = LightDarkBoxAnalyzer(data_dir='C:\troubleshooting\two_black_animals_14bp\dlc_test', save_path="C:\troubleshooting\two_black_animals_14bp\light_dark_ex\light_dark_data.csv", body_part='Nose_1', fps=30.2)
    >>> analyzer = LightDarkBoxAnalyzer(data_dir='D:\light_dark_box\project_folder\csv\input_csv\test', save_path="C:\troubleshooting\two_black_animals_14bp\light_dark_ex\light_dark_data.csv", body_part='nose', fps=29)
    >>> analyzer.run()
    >>> analyzer.save()

    :references:
       .. [1] For discussion about the development, see - `GitHub issue 446 <https://github.com/sgoldenlab/simba/issues/446#issuecomment-2930692735>`_.
    """


    def __init__(self,
                 data_dir: Union[str, os.PathLike],
                 body_part: str,
                 fps: Union[int, float],
                 threshold: float = 0.01,
                 minimum_episode_duration: float = 10e-16,
                 save_path: Optional[Union[str, os.PathLike]] = None):

        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'], raise_error=True, as_dict=True)
        if save_path is not None:
            check_if_dir_exists(in_dir=os.path.dirname(save_path))
        check_str(name=f'{self.__class__.__name__} body_part', value=body_part)
        check_float(name=f'{self.__class__.__name__} fps', value=fps, min_value=10e-6)
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, min_value=0.0, max_value=1.0)
        check_float(name=f'{self.__class__.__name__} minimum_duration', value=minimum_episode_duration, min_value=0.0)
        self.bp_cols = [f'{body_part}_x', f'{body_part}_y', f'{body_part}_p']
        self.save_path, self.body_part, self.fps, self.threshold, self.min_dur = save_path, body_part, fps, threshold, minimum_episode_duration
        self.file_cnt = len(list(self.data_paths.keys()))

    def _remove_outliers(self, df: pd.DataFrame):
        outlier_sequences = [g for g in (list(map(lambda x: x[1], grp)) for _, grp in groupby(enumerate(df.index[df['DURATION (S)'] < self.min_dur]), lambda x: x[0] - x[1]))]
        for outlier_sequence in outlier_sequences:
            start = df.loc[outlier_sequence[0]] if 0 in outlier_sequence else df.loc[outlier_sequence[0] - 1]
            end = df.loc[outlier_sequence[-1]]
            start['END TIME (S)'] = end['END TIME (S)']
            start['END FRAME'] = end['END FRAME']
            start['DURATION (S)'] = ((start['END FRAME'] - start['START FRAME']) + 1) / self.fps
            df = df.drop(index=outlier_sequence)
            df.loc[start.name] = start
        df['group'] = (df['EVENT'] != df['EVENT'].shift()).cumsum()
        df = (df.groupby(['group', 'EVENT', 'VIDEO', 'BODY-PART'], as_index=False).agg({'START TIME (S)': 'first', 'END TIME (S)': 'last', 'START FRAME': 'first', 'END FRAME': 'last', 'DURATION (S)': 'sum'})).drop(columns='group')
        df = df[OUT_COL_NAMES]
        df['START TIME (HH:MM:SS)'] = seconds_to_timestamp(seconds=list(df['START TIME (S)']))
        df['END TIME (HH:MM:SS)'] = seconds_to_timestamp(seconds=list(df['END TIME (S)']))
        return df

    def run(self):
        self.results = []
        for file_cnt, (file_name, file_path) in enumerate(self.data_paths.items()):
            video_timer = SimbaTimer(start=True)
            self.data_df = pd.read_csv(file_path).reset_index(drop=True).iloc[:, 1:]
            headers = self.data_df.iloc[0].unique()
            cols = np.array([np.array([f'{x}_x', f'{x}_y', f'{x}_p']) for x in headers]).flatten()
            self.data_df = self.data_df.iloc[2:].reset_index(drop=True).apply(pd.to_numeric, errors='coerce').fillna(0)
            self.data_df.columns = cols
            check_valid_dataframe(df=self.data_df, valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.bp_cols)
            self.bp_p = self.data_df[self.bp_cols[2]]
            self.data_df[self.bp_cols[2]] = (self.data_df[self.bp_cols[2]] >= self.threshold).astype(int)
            light_bouts = detect_bouts(data_df=self.data_df, target_lst=[self.bp_cols[2]], fps=self.fps)
            self.data_df[self.bp_cols[2]] = 1 - self.data_df[self.bp_cols[2]]
            dark_bouts = detect_bouts(data_df=self.data_df, target_lst=[self.bp_cols[2]], fps=self.fps)
            dark_bouts.columns, light_bouts.columns = COLUMN_NAMES, COLUMN_NAMES
            light_bouts['EVENT'], dark_bouts['EVENT'] = 'LIGHT BOX', 'DARK BOX'
            light_bouts['BODY-PART'], dark_bouts['BODY-PART'] = self.body_part, self.body_part
            light_bouts['VIDEO'], dark_bouts['VIDEO'] = file_name, file_name
            light_bouts, dark_bouts = light_bouts[OUT_COL_NAMES], dark_bouts[OUT_COL_NAMES]
            out = pd.concat([light_bouts, dark_bouts], axis=0).reset_index(drop=True).sort_values(by=['VIDEO', 'START TIME (S)']).reset_index(drop=True)
            out = self._remove_outliers(df=out)
            self.results.append(out)
            video_timer.stop_timer()
            print(f'Light-dark analysis complete video {file_name} ({file_cnt+1}/{self.file_cnt})... (elapsed time: {video_timer.elapsed_time_str}s)')
        self.results = pd.concat(self.results, axis=0)
        self.results = self.results.sort_values(by=['VIDEO', 'START TIME (S)']).reset_index(drop=True)


    def save(self):
        write_df(df=self.results, file_type='csv', save_path=self.save_path)
        print(f'COMPLETE: Light-dark box data for {self.file_cnt} video(s) saved at {self.save_path}')


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, required=True, help='Path to directory containing DeepLabCut CSV files.')
    parser.add_argument('--save_path', type=str, required=True, help='Path to save the output CSV file with light/dark bout data.')
    parser.add_argument('--body_part', type=str, required=True, help='Name of the body part to use for determining visibility (e.g., "snout", "center").')
    parser.add_argument('--fps', type=float, required=True, help='Frames per second (int or float)')
    parser.add_argument('--threshold', type=float, required=True, help='Deeplabcut probability value. If below this value, animal is in dark box. If above, animal is in light box', default=0.1)
    parser.add_argument('--minimum_episode_duration', type=float, required=True, help='The shortest allowed visit to either the dark or the light box. Used to remove spurrious errors in the DLC tracking.', default=1)
    args = parser.parse_args()

    detector = LightDarkBoxAnalyzer(data_dir=args.data_dir,
                                    save_path=args.save_path,
                                    body_part=args.body_part,
                                    fps=args.fps,
                                    threshold=args.threshold,
                                    minimum_episode_duration=args.minimum_episode_duration)

    detector.run()
    detector.save()







__author__ = "Simon Nilsson"

import multiprocessing
import os
import platform
import threading
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (DropDownMenu, Entry_Box, FileSelect,
                                        FolderSelect)
from simba.unsupervised.cluster_video_visualizer import ClusterVideoVisualizer
from simba.unsupervised.enums import UMLOptions
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Formats
from simba.utils.lookups import get_color_dict


class ClusterVisualizerPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="CLUSTER VIDEO VISUALIZATIONS")
        ConfigReader.__init__(self, config_path=config_path)
        self.bg_clr_dict = get_color_dict()
        self.include_pose_var = BooleanVar(value=False)
        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        plt_types = ["SKELETON", "HULL", "POINTS"]
        self.plot_type_dropdown = DropDownMenu(
            self.data_frm,
            "PLOT TYPE:",
            plt_types,
            "25",
            com=lambda x: self._activate_video(selection=x),
        )
        self.plot_type_dropdown.setChoices("SKELETON")
        self.dataset_file_selected = FileSelect(
            self.data_frm,
            "CLUSTER DATASET (PICKLE): ",
            lblwidth=25,
            initialdir=self.project_path,
            file_types=[("SimBA model", f"*.{Formats.PICKLE.value}")],
        )
        self.plot_type_dropdown.grid(row=0, column=0, sticky=NW)
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.dataset_file_selected.grid(row=1, column=0, sticky=NW)
        self.settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )

        max_vid_lst = [str(x) for x in range(0, 105, 5)]
        max_vid_lst.insert(0, "None")
        self.max_vid_per_cluster = DropDownMenu(
            self.settings_frm, "MAX VIDEOS PER CLUSTER:", max_vid_lst, "25"
        )
        self.max_vid_per_cluster.setChoices(max_vid_lst[0])
        self.speed_dropdown = DropDownMenu(
            self.settings_frm, "VIDEO SPEED:", UMLOptions.SPEED_OPTIONS.value, "25"
        )
        self.bg_clr_dropdown = DropDownMenu(
            self.settings_frm, "BACKGROUND COLOR:", list(self.bg_clr_dict.keys()), "25"
        )
        self.bg_clr_dropdown.setChoices("White")
        self.speed_dropdown.setChoices(1.0)
        self.settings_frm.grid(row=1, column=0, sticky=NW)
        self.max_vid_per_cluster.grid(row=0, column=0, sticky=NW)
        self.speed_dropdown.grid(row=1, column=0, sticky=NW)
        self.bg_clr_dropdown.grid(row=2, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def _activate_video(self, selection):
        if selection == "VIDEO":
            self.videos_dir_select = FolderSelect(
                self.data_frm,
                "VIDEOS DIRECTORY:",
                lblwidth=15,
                initialdir=self.video_dir,
            )
            self.videos_dir_select.grid(row=0, column=1, sticky=NW)
        else:
            self.videos_dir_select.destroy()

    def run(self):
        check_file_exist_and_readable(file_path=self.dataset_file_selected.file_path)
        speed = float(self.speed_dropdown.getChoices())
        data_path = self.dataset_file_selected.file_path
        max_videos = self.max_vid_per_cluster.getChoices()
        if max_videos == "None":
            max_videos = None
        else:
            max_videos = int(max_videos)
        plt_type = self.plot_type_dropdown.getChoices()
        bg_clr = self.color_dict[self.bg_clr_dropdown.getChoices()]

        cluster_visualizer = ClusterVideoVisualizer(
            config_path=self.config_path,
            data_path=data_path,
            plot_type=plt_type,
            bg_clr=bg_clr,
            max_videos=max_videos,
            speed=speed,
        )
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        threading.Thread(target=cluster_visualizer.run()).start()


# _ = ClusterVisualizerPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import threading
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import DropDownMenu, FolderSelect
from simba.unsupervised.clusterer_comparison_calculator import \
    ClustererComparisonCalculator
from simba.utils.checks import check_if_dir_exists
from simba.utils.enums import Formats, Options
from simba.utils.errors import CountError

ADJ_MUTUAL_INFO = "ADJUSTED MUTUAL INFORMATION"
FOWLKES_MALLOWS = "FOWLKES MALLOWS"
ADJ_RAND_INDEX = "ADJUSTED RAND INDEX"


class ClustererComparisonPopUp(PopUpMixin, ConfigReader):
    """
    Pop up for computing clusterer comparisons.

    :example:
    >>> _ = ClustererComparisonPopUp(config_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini')
    """

    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="DATA EXTRACTOR")
        ConfigReader.__init__(self, config_path=config_path)
        data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.data_dir_select = FolderSelect(
            data_frm,
            "DATA DIRECTORY (PICKLES):",
            lblwidth=25,
            initialdir=self.project_path,
        )
        data_frm.grid(row=0, column=0, sticky=NW)
        self.data_dir_select.grid(row=0, column=0, sticky=NW)
        self.cb_dict = self.create_cb_frame(
            main_frm=self.main_frm,
            cb_titles=[ADJ_MUTUAL_INFO, FOWLKES_MALLOWS, ADJ_RAND_INDEX],
            frm_title="STATISTICS",
        )
        self.create_run_frm(title="RUN", run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        stats_selections = []
        check_if_dir_exists(in_dir=self.data_dir_select.folder_path)
        for k, v in self.cb_dict.items():
            if v.get():
                stats_selections.append(k)
        if len(stats_selections) == 0:
            raise CountError(
                msg="Select at least ONE measurement.", source=self.__class__.__name__
            )
        cluster_comparer = ClustererComparisonCalculator(
            config_path=self.config_path,
            data_dir=self.data_dir_select.folder_path,
            statistics=[x.lower() for x in stats_selections],
        )
        cluster_comparer.run()


# ClustererComparisonPopUp(config_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import os
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from simba.data_processors.timebins_movement_calculator import \
    TimeBinsMovementCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_if_valid_rgb_tuple, check_int,
    check_str, check_valid_boolean, check_valid_lst, check_valid_tuple)
from simba.utils.enums import Formats
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success

VELOCITY = 'Velocity (cm/s)'
MEASUREMENT = 'MEASUREMENT'
VALUE = 'VALUE'
MOVEMENT = 'Movement (cm)'
TIME_BIN = 'TIME BIN #'

class DataPlotter(ConfigReader):
    """
    Tabular data visualization of animal movement and distances in the current frame and their aggregate
    statistics.

    :param Union[str, os.PathLike] config_path:  Path to the SimBA project config file in ConfigParser format.
    :param List[Tuple[str, Tuple[int, int, int]]] body_parts:   A list of tuples, where each tuple consists of a body-part name (str) and an RGB tuple (int, int, int) indicating the text color used for that body-part in plots.
    :param List[Union[str, os.PathLike]] data_paths: List of paths to the CSV files containing time-binned animal movement data.
    :param Tuple[int, int, int] bg_clr: Background color of the output image(s) as an RGB tuple. Default is white (255, 255, 255).
    :param Tuple[int, int, int] header_clr:  Text color for the header labels (e.g., "ANIMAL", "TOTAL MOVEMENT") as an RGB tuple. Default is black (0, 0, 0).
    :param int font_thickness:  Thickness of the font used in output images. Must be >= 1. Default is 2.
    :param Tuple[int, int] img_size: Size of the output image as a tuple (width, height). Default is (640, 480).
    :param int decimals: Number of decimal places to round movement and velocity values. Must be >= 0. Default is 2.
    :param bool video_setting: Whether to generate a video output of the data plot. At least one of `video_setting` or `frame_setting` must be True.
    :param bool frame_setting: Whether to generate individual frame image outputs for each time bin.
    :param bool verbose: Whether to print progress information during execution. Default is True.

    .. note::
        `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-data-tables`>_.

    .. image:: _static/img/data_plot.png
       :width: 300
       :align: center

    :examples:
    >>> _ = DataPlotter(config_path='MyConfigPath').run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 body_parts: List[Tuple[str, Tuple[int, int, int]]],
                 data_paths: List[Union[str, os.PathLike]],
                 bg_clr: Tuple[int, int, int] = (255, 255, 255),
                 header_clr: Tuple[int, int, int] = (0, 0, 0),
                 font_thickness: int = 2,
                 img_size: Tuple[int, int] = (640, 480),
                 decimals: int = 2,
                 video_setting: bool = True,
                 frame_setting: bool = False,
                 verbose: bool = True):

        if (not video_setting) and (not frame_setting):
            raise NoSpecifiedOutputError(msg="SIMBA ERROR: Please choose to create video and/or frames data plots. SimBA found that you ticked neither video and/or frames", source=self.__class__.__name__)
        check_valid_lst(data=data_paths, source=self.__class__.__name__, valid_dtypes=(str,), min_len=1)
        for data_path in data_paths: check_file_exist_and_readable(file_path=data_path)
        check_valid_lst(data=body_parts, source=self.__class__.__name__, valid_dtypes=(tuple,))
        ConfigReader.__init__(self, config_path=config_path)
        for bp_cnt, bp_attr in enumerate(body_parts):
            check_str(name=f'{self.__class__.__name__} body-part {bp_cnt}', value=bp_attr[0], options=self.body_parts_lst)
            check_if_valid_rgb_tuple(data=bp_attr[1], raise_error=True, source=f'{self.__class__.__name__} body-part {bp_cnt}')
        self.bp_names, self.txt_clrs = [i[0] for i in body_parts], [i[1] for i in body_parts]
        check_if_valid_rgb_tuple(data=bg_clr, raise_error=True, source=f'{self.__class__.__name__} bg_clr')
        check_if_valid_rgb_tuple(data=header_clr, raise_error=True, source=f'{self.__class__.__name__} header_clr')
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        check_int(name=f'{self.__class__.__name__} font_thickness', value=font_thickness, min_value=1)
        check_valid_tuple(x=img_size, source=f'{self.__class__.__name__} img_size', accepted_lengths=(2,), valid_dtypes=(int,), min_integer=1)
        check_int(name=f'{self.__class__.__name__} decimals', value=decimals, min_value=0)
        self.video_setting, self.frame_setting, self.body_parts = video_setting, frame_setting, body_parts
        self.data_paths, self.bg_clr = data_paths, bg_clr
        self.font_thickness, self.img_size, self.decimals, self.header_clr, self.verbose = font_thickness, img_size, decimals, header_clr, verbose
        self.font_size, self.col_shift, self.row_shift = PlottingMixin().get_optimal_font_scales(text='TOTAL MOVEMENT (CM) + EAT', accepted_px_width=int(img_size[0]/3), accepted_px_height=int(img_size[1]/3), text_thickness=font_thickness)
        if not os.path.exists(self.data_table_path):
            os.makedirs(self.data_table_path)
        print(f"Processing {len(self.data_paths)} video(s)...")
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        self.timebins_calculator = TimeBinsMovementCalculator(config_path=self.config_path, bin_length=1, file_paths=data_paths, plots=False, verbose=True, body_parts=self.bp_names)
        self.timebins_calculator.run()
        self.__compute_spacings()


    def __compute_spacings(self):
        """
        Private helper to compute appropriate spacing between printed text.
        """
        self.loc_dict = {}
        self.loc_dict["Animal"] = (10, self.row_shift)
        self.loc_dict["total_movement_header"] = (10+self.col_shift, self.row_shift)
        self.loc_dict["current_velocity_header"] = (int(10+(self.col_shift*2)), self.row_shift)
        self.loc_dict["animals"] = {}
        for animal_cnt, animal_name in enumerate(self.body_parts):
            self.loc_dict["animals"][animal_name[0]] = {"index_loc":           (10, int(self.row_shift * (animal_cnt +2))),
                                                        'total_movement_loc':  (10+self.col_shift, int(self.row_shift * (animal_cnt +2))),
                                                        'current_velocity_loc': (int(10+(self.col_shift*2)), int(self.row_shift * (animal_cnt +2)))}

    def run(self):
        def multiprocess_img_creation(data: list,
                                      location_dict: dict,
                                      video_data: pd.DataFrame,
                                      bg_clr: Tuple[int, int, int],
                                      text_clrs: List[Tuple[int, int, int]],
                                      font_size: int,
                                      header_clr: Tuple[int, int, int],
                                      font_thickness: int,
                                      img_size: tuple,
                                      decimals: int):

            time_bin_data = video_data[video_data['TIME BIN #'] == data[0]]
            img = np.zeros((img_size[1], img_size[0], 3))
            img[:] = bg_clr


            img = PlottingMixin().put_text(img=img, text='ANIMAL', pos=location_dict["Animal"], font_size=self.font_size, font_thickness=font_thickness, text_color=header_clr, text_bg_alpha=1, text_color_bg=bg_clr, font=cv2.FONT_HERSHEY_TRIPLEX)
            img = PlottingMixin().put_text(img=img, text='TOTAL MOVEMENT (CM)', pos=location_dict["total_movement_header"], font_size=self.font_size, font_thickness=font_thickness, text_color=header_clr, text_bg_alpha=1, text_color_bg=bg_clr, font=cv2.FONT_HERSHEY_TRIPLEX)
            img = PlottingMixin().put_text(img=img, text="VELOCITY (CM/S)", pos=location_dict["current_velocity_header"], font_size=font_size, font_thickness=font_thickness, text_color=header_clr, text_bg_alpha=1, text_color_bg=bg_clr, font=cv2.FONT_HERSHEY_TRIPLEX)

            for bp_cnt, bp_name in enumerate(time_bin_data['BODY-PART'].unique()):
                velocity_data = time_bin_data[time_bin_data['BODY-PART'] == bp_name]
                bp_data = video_data[video_data['BODY-PART'] == bp_name]
                velocity_data = round(velocity_data[velocity_data[MEASUREMENT] == VELOCITY]['VALUE'].values[0], decimals)
                movement_data = bp_data[bp_data[MEASUREMENT] == MOVEMENT]
                movement_data = round(movement_data[movement_data[TIME_BIN] <= data[0]][VALUE].sum(), decimals)
                img = PlottingMixin().put_text(img=img, text=bp_name, pos=location_dict["animals"][bp_name]["index_loc"], font_size=font_size, font_thickness=font_thickness, text_color=text_clrs[bp_cnt], text_bg_alpha=1, text_color_bg=bg_clr, font=cv2.FONT_HERSHEY_TRIPLEX)
                img = PlottingMixin().put_text(img=img, text=str(movement_data), pos=location_dict["animals"][bp_name]["total_movement_loc"], font_size=font_size, font_thickness=font_thickness, text_color=text_clrs[bp_cnt], text_bg_alpha=1,font=cv2.FONT_HERSHEY_TRIPLEX, text_color_bg=bg_clr)
                img = PlottingMixin().put_text(img=img, text=str(velocity_data), pos=location_dict["animals"][bp_name]["current_velocity_loc"], font_size=font_size, font_thickness=font_thickness, text_color=text_clrs[bp_cnt], text_bg_alpha=1,font=cv2.FONT_HERSHEY_TRIPLEX, text_color_bg=bg_clr)

            return img


        for video_cnt, video_data in enumerate(self.timebins_calculator.out_df_lst):
            video_timer = SimbaTimer(start=True)
            video_name = video_data['VIDEO'].loc[0]
            _, _, self.fps = self.read_video_info(video_name=video_name)
            if self.video_setting:
                self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
                self.video_save_path = os.path.join(self.data_table_path, f"{video_name}.mp4")
                self.writer = cv2.VideoWriter(self.video_save_path, self.fourcc, self.fps, self.img_size)
            if self.frame_setting:
                self.frame_save_path = os.path.join(self.data_table_path, video_name)
                if not os.path.exists(self.frame_save_path): os.makedirs(self.frame_save_path)


            time_bins = [[x] for x in list(set(video_data['TIME BIN #']))]
            print('Creating images...')
            self.imgs = Parallel(n_jobs=self.cpu_to_use, verbose=1, backend="threading")(delayed(multiprocess_img_creation)(x,
                                                                                                                            self.loc_dict,
                                                                                                                            video_data,
                                                                                                                            self.bg_clr,
                                                                                                                            self.txt_clrs,
                                                                                                                            self.font_size,
                                                                                                                            self.header_clr,
                                                                                                                            self.font_thickness,
                                                                                                                            self.img_size,
                                                                                                                            self.decimals) for x in time_bins)
            frm_cnt = 0
            for img_cnt, img in enumerate(self.imgs):
                for frame_cnt in range(int(self.fps)):
                    if self.video_setting:
                        self.writer.write(np.uint8(img))
                    if self.frame_setting:
                        frm_save_name = os.path.join(self.frame_save_path, f"{frame_cnt}.png")
                        cv2.imwrite(frm_save_name, np.uint8(img))
                frm_cnt += 1
                if self.verbose:
                    print(f"Frame: {int((img_cnt+1) * self.fps)} / {int(len(self.imgs) * self.fps)}. Video: {video_name} ({video_cnt+1}/{len(self.timebins_calculator.out_df_lst)})")

            print(f"Data tables created for video {video_name}...")
            if self.video_setting:
                self.writer.release()
                video_timer.stop_timer()
                if self.verbose:
                    stdout_success(msg=f'Video {self.video_save_path} complete', source=self.__class__.__name__, elapsed_time=video_timer.elapsed_time_str)

        self.timer.stop_timer()
        if self.verbose:
            stdout_success(msg=f"All data table videos created inside {self.data_table_path}", elapsed_time=self.timer.elapsed_time_str)

#
#
# # style_attr = {'bg_color': 'White', 'header_color': 'Black', 'font_thickness': 1, 'size': (640, 480), 'data_accuracy': 2}
# body_part_attr = [('Tail_base', (0, 255, 0)), ('Nose', (255, 255, 0))]
# data_paths = [r"C:\troubleshooting\mitra\project_folder\csv\outlier_corrected_movement_location\501_MA142_Gi_Saline_0513.csv"]
# #
# #
# test = DataPlotter(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                    body_parts=body_part_attr,
#                    data_paths=data_paths,
#                    video_setting=True,
#                    frame_setting=False)
# test.run()

# style_attr = {'bg_color': 'White', 'header_color': 'Black', 'font_thickness': 1, 'size': (640, 480), 'data_accuracy': 2}
# body_part_attr = [['Ear_left_1', 'Grey'], ['Ear_right_2', 'Red']]
# data_paths = ['/Users/simon/Desktop/envs/simba_dev/tests/test_data/two_C57_madlc/project_folder/csv/outlier_corrected_movement_location/Together_1.csv']
#
#
# test = DataPlotter(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/two_C57_madlc/project_folder/project_config.ini',
#                    style_attr=style_attr,
#                    body_part_attr=body_part_attr,
#                    data_paths=data_paths,
#                    video_setting=True,
#                    frame_setting=False)
# test.create_data_plots()
